#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import mock
import pytest
import os
import asyncio
import sqlite3
import contextlib


import octobot_commons.asyncio_tools as asyncio_tools
import octobot_commons.errors as errors
import octobot_commons.databases as databases
import octobot_commons.enums as enums

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio

DATA_FILE1 = "ExchangeHistoryDataCollector_1589740606.4862757.data"
DATA_FILE2 = "second_ExchangeHistoryDataCollector_1589740606.4862757.data"
OHLCV = mock.Mock(value="ohlcv")
KLINE = mock.Mock(value="kline")


# use context manager instead of fixture to prevent pytest threads issues
@contextlib.asynccontextmanager
async def get_database(data_file=DATA_FILE1):
    async with databases.new_sqlite_database(os.path.join("tests", "static", data_file)) as db:
        yield db
    # prevent "generator didn't stop after athrow(), see https://github.com/python-trio/trio/issues/2081"
    await asyncio_tools.wait_asyncio_next_cycle()


# use context manager instead of fixture to prevent pytest threads issues
@contextlib.asynccontextmanager
async def get_temp_empty_database():
    database_name = "temp_empty_database"
    try:
        async with databases.new_sqlite_database(database_name) as db:
            yield db
    finally:
        # prevent "generator didn't stop after athrow(), see https://github.com/python-trio/trio/issues/2081"
        await asyncio_tools.wait_asyncio_next_cycle()
        os.remove(database_name)


async def test_invalid_file():
    file_name = "plop"
    db = databases.SQLiteDatabase(file_name)
    try:
        await db.initialize()
        assert not await db.check_table_exists(KLINE)
        with pytest.raises(sqlite3.OperationalError):
            await db.check_table_not_empty(KLINE)
    finally:
        await db.stop()
        os.remove(file_name)


async def test_select():
    async with get_database() as database:
        # default values
        with pytest.raises(errors.DatabaseNotFoundError):
            await database.select(KLINE)

        ohlcv = await database.select(OHLCV)
        assert len(ohlcv) == 6531

        ohlcv = await database.select(OHLCV, time_frame="1h")
        assert len(ohlcv) == 500

        ohlcv = await database.select(OHLCV, symbol="xyz")
        assert len(ohlcv) == 0

        ohlcv = await database.select(OHLCV, symbol="ETH/BTC")
        assert len(ohlcv) == 6531

        changed_order_ohlcv = await database.select(OHLCV, order_by="time_frame", symbol="ETH/BTC")
        assert changed_order_ohlcv[0] != ohlcv[0]

        ohlcv = await database.select(OHLCV, xyz="xyz")
        assert len(ohlcv) == 0


async def test_select_max():
    async with get_database() as database:
        assert await database.select_max(OHLCV, ["timestamp"]) == [(1590883200,)]
        assert await database.select_max(OHLCV, ["timestamp"], time_frame="1h") == [(1589742000,)]
        assert await database.select_max(OHLCV, ["timestamp"], ["symbol"], time_frame="1h") == \
            [(1589742000, "ETH/BTC")]


async def test_select_min():
    async with get_database() as database:
        assert await database.select_min(OHLCV, ["timestamp"]) == [(1500249600,)]
        assert await database.select_min(OHLCV, ["timestamp"], time_frame="1h") == [(1587945600,)]
        assert await database.select_min(OHLCV, ["timestamp"], ["symbol"], time_frame="1h") == \
            [(1587945600, "ETH/BTC")]


async def test_select_count():
    async with get_database() as database:
        assert await database.select_count(OHLCV, ["*"]) == [(6531,)]
        assert await database.select_count(OHLCV, ["*"], time_frame="1h") == [(500,)]
        assert await database.select_count(OHLCV, ["*"], time_frame="1M") == [(35,)]


async def test_select_from_timestamp():
    async with get_database() as database:
        operations = [enums.DataBaseOperations.INF_EQUALS.value]
        candles = await database.select_from_timestamp(OHLCV, ["1587960000"], operations)
        assert len(candles) > 0
        assert all(candle[0] <= 1587960000 for candle in candles)

        operations = [enums.DataBaseOperations.INF_EQUALS.value, enums.DataBaseOperations.SUP_EQUALS.value]
        candles = await database.select_from_timestamp(OHLCV,
                                                       ["1587960000", "1587960000"],
                                                       operations)
        assert len(candles) > 0
        assert all(candle[0] == 1587960000 for candle in candles)

        operations = [enums.DataBaseOperations.INF_EQUALS.value, enums.DataBaseOperations.SUP_EQUALS.value]
        candles = await database.select_from_timestamp(OHLCV,
                                                       ["1587960000", "1587945600"],
                                                       operations)
        assert len(candles) == 15
        assert all(1587945600 <= candle[0] <= 1587960000 for candle in candles)

        operations = [enums.DataBaseOperations.INF_EQUALS.value, enums.DataBaseOperations.SUP_EQUALS.value]
        candles = await database.select_from_timestamp(OHLCV,
                                                       ["1587960000", "1587945600"],
                                                       operations,
                                                       symbol="xyz")
        assert len(candles) == 0


async def test_gather_concurrent_select():
    async with get_database() as database:
        timestamps_1h = [ohlcv[0] for ohlcv in await database.select(OHLCV, time_frame="1h")]
        timestamps_4h = [ohlcv[0] for ohlcv in await database.select(OHLCV, time_frame="4h")]
        coros = [_check_select_result(database, ts, "1h") for ts in timestamps_1h]
        coros += [_check_select_result(database, ts, "4h") for ts in timestamps_4h]
        await asyncio.gather(*coros)


async def test_create_tasks_concurrent_selects():
    async with get_database() as database:
        timestamps_1h = [ohlcv[0] for ohlcv in await database.select(OHLCV, time_frame="1h")]
        timestamps_1m = [ohlcv[0] for ohlcv in await database.select(OHLCV, time_frame="1m")]
        timestamps_4h = [ohlcv[0] for ohlcv in await database.select(OHLCV, time_frame="4h",
                                                                     size=50)]

        calls_count = len(timestamps_1h) + len(timestamps_4h) + len(timestamps_1m)
        failed_calls = []
        success_calls = []

        async def select_task(db, timestamp, time_frame):
            try:
                await _check_select_result(db, timestamp, time_frame)
                success_calls.append((timestamp, time_frame))
            except Exception as e:
                failed_calls.append((timestamp, time_frame, e))

        tasks = []
        for ts in timestamps_1h:
            tasks.append(asyncio.get_event_loop().create_task(select_task(database, ts, "1h")))
        for ts in timestamps_4h:
            tasks.append(asyncio.get_event_loop().create_task(select_task(database, ts, "4h")))
        for ts in timestamps_1m:
            tasks.append(asyncio.get_event_loop().create_task(select_task(database, ts, "1m")))
            # for wait for next cycle to make previous requests end and re-use previous cursors
            await asyncio_tools.wait_asyncio_next_cycle()

        await asyncio.gather(*tasks)
        assert len(success_calls) == calls_count
        assert failed_calls == []


async def test_stop_while_concurrent_select():
    async with get_database() as database:
        timestamps = [ohlcv[0] for ohlcv in await database.select(OHLCV, time_frame="1h")]
        await _check_select_result(database, timestamps[0])
        asyncio.create_task(asyncio.wait(
            asyncio.gather(*[_check_select_result(database, ts, expected_exception=sqlite3.ProgrammingError)
                             for ts in timestamps])))
        # not enough time to finish all requests, most if not all will remaining pending
        await asyncio_tools.wait_asyncio_next_cycle()


async def test_double_database():
    async with get_database() as database1, get_database(DATA_FILE2) as database2:
        timestamps1 = [ohlcv[0] for ohlcv in await database1.select(OHLCV, time_frame="1h")]
        timestamps2 = [ohlcv[0] for ohlcv in await database2.select(OHLCV, time_frame="1h")]
        await asyncio.gather(*[_check_select_result(database1, ts) for ts in timestamps1])
        await asyncio.gather(*[_check_select_result(database2, ts) for ts in timestamps2])


async def test_double_database_stop_while_concurrent_select():
    async with get_database() as database1, get_database(DATA_FILE2) as database2:
        timestamps1 = [ohlcv[0] for ohlcv in await database1.select(OHLCV, time_frame="1h")]
        timestamps2 = [ohlcv[0] for ohlcv in await database2.select(OHLCV, time_frame="1h")]
        await _check_select_result(database1, timestamps1[0])
        await _check_select_result(database2, timestamps2[0])
        asyncio.create_task(asyncio.wait(
            asyncio.gather(*[_check_select_result(database1, ts, expected_exception=sqlite3.ProgrammingError)
                             for ts in timestamps1])))
        asyncio.create_task(asyncio.wait(
            asyncio.gather(*[_check_select_result(database2, ts, expected_exception=sqlite3.ProgrammingError)
                             for ts in timestamps2])))
        # not enough time to finish all requests, most if not all will remaining pending
        await asyncio_tools.wait_asyncio_next_cycle()


async def test_insert():
    async with get_temp_empty_database() as temp_empty_database:
        await temp_empty_database.insert(OHLCV, symbol="xyz", timestamp=1, price=1, date="01")
        assert await temp_empty_database.select(OHLCV) == [(1, 'xyz', '1', '01')]


async def test_insert_all():
    async with get_temp_empty_database() as temp_empty_database:
        await temp_empty_database.insert_all(OHLCV,
                                             symbol=["xyz", "abc"],
                                             timestamp=[1, 2],
                                             price=[1, 10],
                                             date=["01", "05"])
        assert await temp_empty_database.select(OHLCV) == [(2, 'abc', '10', '05'), (1, 'xyz', '1', '01')]
        assert await temp_empty_database.select(OHLCV, date="05") == [(2, 'abc', '10', '05')]


async def test_create_index():
    async with get_temp_empty_database() as temp_empty_database:
        await temp_empty_database.insert(OHLCV, 1, symbol="xyz", price="1", date="01")
        # ensure no exception
        await temp_empty_database.create_index(OHLCV, ["symbol", "timestamp"])
        assert await temp_empty_database.select(OHLCV) == [(1, 'xyz', '1', '01')]


async def _check_select_result(database, timestamp, time_frame="1h", expected_exception=None):
    try:
        ohlcv = await database.select(OHLCV, time_frame=time_frame, timestamp=str(timestamp))
        assert len(ohlcv) == 1
        assert ohlcv[0][0] == timestamp
    except Exception as e:
        if e.__class__ is expected_exception:
            pass
        else:
            raise



# pylint: disable=C0116
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.enums


class SignalBundle:
    def __init__(self, identifier: str, signals=None, version=None):
        self.identifier: str = identifier
        self.signals: list = signals or []
        self.version: str = version or self._get_version()

    def to_dict(self) -> dict:
        return {
            octobot_commons.enums.SignalBundlesAttrs.IDENTIFIER.value: self.identifier,
            octobot_commons.enums.SignalBundlesAttrs.SIGNALS.value: [
                signal.to_dict() for signal in self.signals
            ],
            octobot_commons.enums.SignalBundlesAttrs.VERSION.value: self.version,
        }

    def __str__(self):
        return f"{self.to_dict()}"

    # pylint: disable=C0415
    def _get_version(self) -> str:
        try:
            import octobot.constants

            return octobot.constants.COMMUNITY_FEED_CURRENT_MINIMUM_VERSION
        except ImportError:
            return "1.0.0"



# pylint: disable=W0718
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import json
import jsonschema
import octobot_commons.logging


LOGGER_NAME = "json_util"


def validate(config, schema_file) -> None:
    """
    Validate a config file, raise upon validation error
    :param config: the config
    :param schema_file: the config schema
    :return: None
    """
    with open(schema_file) as json_schema:
        loaded_schema = json.load(json_schema)
    jsonschema.validate(instance=config, schema=loaded_schema)


def read_file(
    file_path: str,
    raise_errors: bool = True,
    on_error_value: dict = None,
    open_mode="r",
) -> dict:
    """
    Read a load the given file with json.load()
    :param file_path: file to read
    :param raise_errors: when True will forward errors. Will just log errors otherwise
    :param on_error_value: return this value when raise_errors is False and an error occurs
    :param open_mode: the file open mode to give to open()
    :return: the parsed file or default value on error if possible
    """
    try:
        with open(file_path, open_mode) as open_file:
            return json.load(open_file)
    except PermissionError as err:
        if raise_errors:
            raise
        octobot_commons.logging.get_logger(LOGGER_NAME).error(
            f"Permission error when reading {file_path} file: {err}."
        )
    except Exception as err:
        if raise_errors:
            raise
        octobot_commons.logging.get_logger(LOGGER_NAME).exception(
            f"Unexpected error when reading {file_path} file: {err}."
        )
    if on_error_value is None:
        raise ValueError("on_error_value is unset")
    return on_error_value



# pylint: disable=R0913,R0801
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.databases.bases.base_database as base_database
import octobot_commons.databases.document_database_adaptors as adaptors
import octobot_commons.errors as commons_errors
import octobot_commons.logging as commons_logging


class DBWriter(base_database.BaseDatabase):
    MAX_ROWS_BUFFER_SIZE = 500

    def __init__(
        self,
        file_path: str,
        database_adaptor=adaptors.TinyDBAdaptor,
        cache_size=None,
        **kwargs,
    ):
        super().__init__(
            file_path,
            database_adaptor=database_adaptor,
            cache_size=cache_size,
            **kwargs,
        )
        self.rows_buffer = {}
        self.rows_buffer_size = self.MAX_ROWS_BUFFER_SIZE

    async def log(self, table_name: str, row: dict, cache=True, rows_buffering=False):
        """
        Write a row into the given table
        :param table_name: name of the table
        :param row: row to write
        :param cache: When True, enables local cache
        :param rows_buffering: When True, defer actual database writing using a row buffer
        """
        if cache:
            try:
                self.cache.register(table_name, row)
            except commons_errors.UncachableValue:
                await self._database.insert(table_name, row)
        if rows_buffering:
            await self._buffer_row(table_name, row)
        else:
            await self._database.insert(table_name, row)

    async def update(self, table_name: str, row: dict, query, uuid=None):
        """
        Updates rows
        :param table_name: table to update data from
        :param row: updated data
        :param query: select query
        :param uuid: uuid to select data (enable faster operations)
        :return: the update result
        """
        return await self._database.update(table_name, row, query, uuid=uuid)

    async def upsert(
        self, table_name: str, row: dict, query, uuid=None, cache_query=None
    ):
        """
        Update or insert a row
        Upsert can be a very slow operation: avoid is as much as possible
        Upsert with uuid is fast though, try to use it when possible
        :param table_name: table to update data from
        :param row: updated data
        :param query: select query
        :param uuid: uuid to select data (enable faster operations)
        :param cache_query: query identifier associated to this row, can be used later on for faster upserts on
        the same row using cache
        :return: the upsert result when operating without cache
        """
        if uuid is not None or cache_query is None:
            return await self._database.upsert(table_name, row, query, uuid=uuid)
        if uuid := self.cache.cached_uuid(table_name, str(cache_query)):
            return await self._database.upsert(table_name, row, query, uuid=uuid)
        if result := self.cache.cached_query(table_name, str(cache_query)):
            result.update(row)
        else:
            await self._buffer_row(
                table_name, row, cache_query=cache_query, cache=False
            )
            self.cache.register(table_name, str(cache_query), result=row)
            return None

    async def update_many(self, table_name: str, update_values: list):
        """
        Updates multiple values at once, doesn't use cache
        """
        return await self._database.update_many(table_name, update_values)

    async def delete(self, table_name: str, dict_query: dict):
        """
        Deletes selected values at once, doesn't use cache
        """
        query = None
        if dict_query:
            if isinstance(dict_query, dict):
                self.cache.delete_from_rows_cache(table_name, dict_query)
                query = await self.search()
                for key, value in dict_query.items():
                    query = query[key] == value
            else:
                query = dict_query
        else:
            self.cache.clear(table_name)
        return await self._database.delete(table_name, query)

    async def delete_all(self, table_name: str):
        """
        Deletes all rows from a table
        """
        self.cache.clear(table_name)
        return await self._database.delete(table_name, None)

    async def log_many(self, table_name: str, rows: list, cache=True):
        """
        Inserts multiple values into the given table
        :param table_name: name of the table
        :param rows: rows to insert
        :param cache: when True, rows are written into a cache buffer and written in bulk when buffer will be full
        """
        if cache:
            for row in rows:
                try:
                    self.cache.register(table_name, row)
                except commons_errors.UncachableValue:
                    # can pass here since row will be inserted anyway
                    pass
        return await self._database.insert_many(table_name, rows)

    async def replace_all(self, table_name, rows: list, cache=True):
        """
        Deletes everything for the give table name and replace the content by the given rows
        :param table_name: name of the table
        :param rows: rows to insert
        :param cache: When True, rows will be registered in cache
        """
        await self.delete_all(table_name)
        await self.log_many(table_name, rows, cache=cache)

    async def flush(self):
        """
        Flushes all caches and "commit" to the database then forces the database to flush its own internal cache if any
        """
        try:
            await self._flush_all_rows_buffers(cache=True)
            await super().flush()
        except TypeError as err:
            commons_logging.get_logger(str(self)).exception(
                err,
                True,
                f"Error when writing database, this is probably due to a script that is "
                f"saving a non json-serializable value: {err}",
            )

    async def _buffer_row(self, table, row, cache_query=None, cache=True):
        try:
            self.rows_buffer[table].append((row, cache_query))
            if len(self.rows_buffer[table]) >= self.rows_buffer_size:
                await self._flush_rows_buffer(table, cache=cache)
        except KeyError:
            self.rows_buffer[table] = [(row, cache_query)]

    async def _flush_rows_buffer(self, table, cache=True):
        uuids = await self.log_many(
            table, tuple(row[0] for row in self.rows_buffer[table]), cache=cache
        )
        for index, row in enumerate(self.rows_buffer[table]):
            self.cache.register(table, str(row[1]), uuid=uuids[index])
        self.rows_buffer[table] = []

    async def _flush_all_rows_buffers(self, cache=True):
        for table, rows in self.rows_buffer.items():
            uuids = await self.log_many(
                table, tuple(row[0] for row in rows), cache=cache
            )
            if cache:
                for index, row in enumerate(rows):
                    self.cache.register(table, str(row[1]), uuid=uuids[index])
            self.rows_buffer[table] = []



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.databases.run_databases.run_databases_provider as run_databases_provider
import octobot_commons.configuration as configuration


async def init_bot_storage(bot_id, run_database_identifier, clear_user_inputs):
    """
    Initializes the associated bot_id databases. Deletes any existing user input if clear_user_inputs is True
    """
    if not run_databases_provider.RunDatabasesProvider.instance().has_bot_id(bot_id):
        # only one run database per bot id
        await run_databases_provider.RunDatabasesProvider.instance().add_bot_id(
            bot_id, run_database_identifier
        )
        if clear_user_inputs:
            await configuration.clear_user_inputs(
                run_databases_provider.RunDatabasesProvider.instance().get_run_db(
                    bot_id
                )
            )


async def close_bot_storage(bot_id):
    """
    :return: Close the bot_id associated run databases
    """
    if run_databases_provider.RunDatabasesProvider.instance().has_bot_id(bot_id):
        await run_databases_provider.RunDatabasesProvider.instance().close(bot_id)



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import contextlib
import octobot_commons.logging


class DocumentDatabase:
    """
    DocumentDatabase is used to communicate with an underlying database
    """

    def __init__(self, database_adaptor):
        """
        DocumentDatabase constructor
        :param database_adaptor: database adaptor
        """
        self.adaptor = database_adaptor

    def initialize(self):
        """
        Initialize the database adaptor.
        """
        self.adaptor.initialize()

    def get_uuid(self, document) -> int:
        """
        Returns the uuid of the document
        :param document: the document
        """
        return self.adaptor.get_uuid(document)

    def get_db_path(self):
        """
        Select database path
        """
        return self.adaptor.db_path

    async def select(self, table_name: str, query, uuid=None) -> list:
        """
        Select data from the table_name table
        :param table_name: name of the table
        :param query: select query
        :param uuid: id of the document
        """
        return await self.adaptor.select(table_name, query, uuid=uuid)

    async def tables(self) -> list:
        """
        Select tables
        """
        return await self.adaptor.tables()

    async def insert(self, table_name: str, row: dict) -> int:
        """
        Insert dict data into the table_name table
        :param table_name: name of the table
        :param row: data to insert
        """
        return await self.adaptor.insert(table_name, row)

    async def upsert(self, table_name: str, row: dict, query, uuid=None) -> int:
        """
        Insert or update dict data into the table_name table
        :param table_name: name of the table
        :param row: data to insert
        :param query: select query
        :param uuid: id of the document
        """
        return await self.adaptor.upsert(table_name, row, query, uuid=uuid)

    async def insert_many(self, table_name: str, rows: list) -> list:
        """
        Insert multiple dict data into the table_name table
        :param table_name: name of the table
        :param rows: data to insert
        """
        return await self.adaptor.insert_many(table_name, rows)

    async def update(self, table_name: str, row: dict, query: dict, uuid=None) -> list:
        """
        Insert dict data into the table_name table
        :param table_name: name of the table
        :param row: data to update
        :param query: select statement
        :param uuid: id of the document
        """
        return await self.adaptor.update(table_name, row, query, uuid=uuid)

    async def update_many(self, table_name: str, update_values: list) -> list:
        """
        Update multiple values from the table_name table
        :param table_name: name of the table
        :param update_values: values to update
        """
        return await self.adaptor.update(table_name, update_values)

    async def delete(self, table_name: str, query, uuid=None) -> list:
        """
        Delete data from the table_name table
        :param table_name: name of the table
        :param query: select query
        :param uuid: id of the document
        """
        return await self.adaptor.delete(table_name, query, uuid=uuid)

    async def count(self, table_name: str, query) -> int:
        """
        Counts documents in the table_name table
        :param table_name: name of the table
        :param query: select query
        """
        return await self.adaptor.count(table_name, query)

    async def query_factory(self):
        """
        Creates a new empty select query
        """
        return await self.adaptor.query_factory()

    async def hard_reset(self):
        """
        Completely reset the database
        """
        self.get_logger().debug("hard resetting database")
        return await self.adaptor.hard_reset()

    async def flush(self):
        """
        Flushes the database cache
        """
        self.get_logger().debug("flushing database")
        return await self.adaptor.flush()

    async def close(self):
        """
        Closes the database
        """
        self.get_logger().debug("closing database")
        return await self.adaptor.close()

    def get_logger(self):
        """
        :return: the database logger
        """
        return octobot_commons.logging.get_logger(str(self))

    def __str__(self):
        return f"{self.__class__.__name__} with adaptor: {self.adaptor}"

    @classmethod
    @contextlib.asynccontextmanager
    async def locked_database(cls, *args, **kwargs):
        """
        Instantiate and then ensure lock is acquired before initializing the database.
        Closes the database and then releases the lock when exiting
        :param args: args to pass to the database constructor
        :param kwargs: kwargs to pass to the database constructor
        """
        instance = None
        lock_acquired = False
        try:
            instance = cls(*args, **kwargs)
            if instance.adaptor.is_multiprocessing():
                await instance.adaptor.acquire()
                lock_acquired = True
            instance.initialize()
            yield instance
        finally:
            if instance is not None:
                try:
                    await instance.close()
                finally:
                    if lock_acquired and instance.adaptor.is_multiprocessing():
                        await instance.adaptor.release()



# cython: language_level=3
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

cdef list _sort_time_frames(list time_frames, bint reverse=*)

cpdef list get_config_time_frame(dict config)
cpdef list sort_time_frames(list time_frames, bint reverse=*)
cpdef void sort_config_time_frames(dict config)
cpdef object get_display_time_frame(dict config, object default_display_time_frame)
cpdef object get_previous_time_frame(list config_time_frames, object time_frame, object origin_time_frame)
cpdef object find_min_time_frame(list time_frames, object min_time_frame=*)
cpdef list parse_time_frames(list time_frames_string_list)
